"""Save all documents."""

from __future__ import annotations

import docxrev

from gradedoc import shared


def save_all():
    """Save all documents without closing them."""

    (paths, _) = shared.get_paths()
    for path in paths:
        document = docxrev.Document(path, save_on_exit=True, close_on_exit=False)
        with document:
            pass



"""Toggle the review pane of the active document."""

from __future__ import annotations

import docxrev
from win32com.client import constants

from gradedoc import shared


def toggle_active_review_pane():
    """Toggle the review pane of the active document."""

    (paths, _) = shared.get_paths()

    active_document = docxrev.get_active_document(save_on_exit=False)
    with active_document:
        # Check if the document is in paths
        in_paths = active_document.path in paths  # we consume `paths` here
        # Now update the grade or raise an exception
        if in_paths:
            active_document.com.ActiveWindow.View.SplitSpecial = (
                constants.wdPaneRevisions
            )
        else:
            raise Exception("Active document not in paths.")



"""Update a document's grade as well as the gradebook."""
from __future__ import annotations

import csv
import os
from dataclasses import dataclass
from typing import Any, Iterator, List

import docxrev
from natsort import natsorted

from gradedoc import shared
from gradedoc.configs import config

__all__ = ["update_grade"]


def update_grade(document: docxrev.Document, gradebook_path: os.PathLike):
    """Update a document's grade as well as the gradebook.

    Parameters
    ----------
    document
        Document.
    gradebook_path
        Gradebook.
    """

    with document:
        grade = grade_document(document)
        update_document_scores(document, grade)
        update_gradebook(document, gradebook_path, grade)


def grade_document(document: docxrev.Document) -> Grade:  # noqa: C901
    """Grade a document.

    Parameters
    ----------
    document
        Document.
    """

    # Search comments in reverse. This means that point deductions in any given section
    # will be encountered before we reach the header comment. This allows us to walk
    # through the comments once, and exit cleanly once we reach the first header
    # comment. We use `iter` and `next` on `comments` because they advance at a
    # different pace than the `header_comment_patterns` and `max_scores`.
    comments = iter(reversed(document.comments))
    header_comment_patterns = reversed(shared.HEADER_COMMENT_PATTERNS)
    max_scores = reversed(shared.MAX_SCORES)

    # Prepare the lists to be returned. They will be reversed at the end.
    scores: List[int] = []
    header_comments: List[docxrev.com.Comment] = []
    deductions = 0
    codes: List[str] = []

    # Get the scores for each section
    comment = safe_next(comments)
    for header_comment_pattern, max_score in zip(header_comment_patterns, max_scores):

        # Process point deductions while we don't recognize a header comment
        content_points_lost = 0
        while not header_comment_pattern.match(comment.text):

            # Increment the points lost in this section for matching comments
            if match := shared.CONTENT_POINTS_LOST_PATTERN.match(comment.text):
                content_points_lost += int(match["value"])

            # Increment total deductions for matching comments
            if match := shared.DEDUCTION_PATTERN.match(comment.text):
                deductions += int(match["value"])

            # Add text description to common deduction codes found in the document
            matches = [
                pattern.match(comment.text)
                for pattern in shared.COMMON_DEDUCTION_PATTERNS
            ]
            if any(matches):
                first_match = [match for match in matches if match][0]
                code = first_match["code"]
                if "D" not in code:  # Skip if this is a custom deduction
                    try:
                        feedback = config.codes[code]
                    except KeyError as error:
                        raise KeyError(
                            f"No entry in configs for code {code}."
                        ) from error
                    first_line_of_comment = comment.text.split("\r", maxsplit=1)[0]
                    comment.update(first_line_of_comment + "\n\n" + feedback)
                    codes.append(code)

            # Try to get the next comment, raising an error if there are none left
            try:
                comment = safe_next(comments)
            except StopIteration as error:
                raise StopIteration(
                    "No comments in document, or template comments are out of order.\n"
                    f"Document: {document.name}"
                ) from error

        # We found a header comment. Store it and move on.
        header_comments.append(comment)
        comment = safe_next(comments)

        # Store the score for this section and skip the header comment
        scores.append(max_score - content_points_lost)

    header_comments.reverse()
    scores.reverse()
    return Grade(header_comments, scores, deductions, "; ".join(natsorted(codes)))  # type: ignore


def update_document_scores(document: docxrev.Document, grade: Grade):
    """Update document scores with the determined grade.

    Parameters
    ----------
    document
        Document.
    grade
        Grade.
    """

    # Update the summary comment scores
    summary_comment = document.comments[0]
    substitution = (
        rf"\g<content>{grade.content}\n"
        rf"\g<deductions>{grade.deductions}\n"
        rf"\g<grade>{grade.total}"
    )
    summary_comment.update(
        shared.SUMMARY_COMMENT_PATTERN.sub(substitution, summary_comment.text)
    )

    # Update the scores of the header comments
    for comment, pattern, score in zip(
        grade.header_comments, shared.HEADER_COMMENT_PATTERNS, grade.scores
    ):
        substitution = rf"\g<header>{score}"
        comment.update(pattern.sub(substitution, comment.text))


def update_gradebook(
    document: docxrev.Document, gradebook_path: os.PathLike, grade: Grade
):
    """Write the grade for the paper being graded to a CSV file.

    Parameters
    ----------
    gradebook_path
        Gradebook.
    document
        Document.
    grade
        Grade.
    """

    # Prepare rows for the CSV
    header_row = [
        "Document",
        "Grade",
        "Total Content",
        "Total Deductions",
        "Deduction Codes",
        *[header.upper() for header in shared.FULL_HEADERS],
    ]

    new_row = [
        document.name,
        grade.total,
        grade.content,
        grade.deductions,
        grade.deduction_codes,
        *grade.scores,
    ]
    rows_to_write: List[Any] = []

    # Create the CSV and write the header if it doesn't exist
    if not os.path.exists(gradebook_path):
        with open(gradebook_path, "w", newline="") as file:
            writer = csv.writer(file)
            writer.writerow(header_row)

    # Build the rows of the CSV, overwriting or appending the current paper being graded
    with open(gradebook_path, "r", newline="") as file:

        update_existing_row = False
        reader = csv.reader(file)
        next(reader)  # skip the header

        # If the paper has already been graded, update that row, otherwise append it
        for row in reader:
            if row and new_row[0] == row[0]:
                rows_to_write.append(new_row)
                update_existing_row = True
            else:
                rows_to_write.append(row)
        if not update_existing_row:
            rows_to_write.append(new_row)

    # Write the updated file
    with open(gradebook_path, "w", newline="") as file:
        writer = csv.writer(file)
        writer.writerow(header_row)
        for row in rows_to_write:
            writer.writerow(row)


def safe_next(comments: Iterator[docxrev.com.Comment]) -> docxrev.com.Comment:
    """Safely get the next comment.

    Safely get the next comment, raising an error if all comments have been exhausted.

    Parameters
    ----------
    comments
        A comment iterator.
    document
        The document.

    Raises
    ------
    StopIteration
        If all comments are exhausted.
    """
    comment = None
    try:
        comment = next(comments)
    except StopIteration as error:
        if comment:
            message = (
                f"Exhausted comments in {comment.in_document.name} "
                "before finding all header comments."
            )
        else:
            message = "No comments in document."
        raise StopIteration(message) from error

    return comment


@dataclass
class Grade:
    """A grade."""

    header_comments: List[docxrev.com.Comment]
    """Header comments."""

    scores: List[int]
    """Scores."""

    deductions: int
    """Deductions."""

    @property
    def content(self) -> int:
        """Content score."""
        return sum(self.scores)

    @property
    def total(self) -> int:
        """Total score."""
        return self.content - self.deductions

    deduction_codes: str
    """ A semicolon delimited string of deduction codes."""



from contextlib import contextmanager
import docxrev
import fire

from gradedoc.add_template_comments import add_template_comments
from gradedoc.close_all import close_all
from gradedoc.copy_example import copy_example
from gradedoc.delete_all_comments import delete_all_comments
from gradedoc.open_all import open_all
from gradedoc.save_all import save_all
from gradedoc.toggle_active_review_pane import toggle_active_review_pane
from gradedoc.update_active_grade import update_active_grade
from gradedoc.update_all_grades import update_all_grades


def main():
    with word():
        fire.Fire(
            {
                "addcom": add_template_comments,
                "example": copy_example,
                "close": close_all,
                "delcom": delete_all_comments,
                "open": open_all,
                "save": save_all,
                "pane": toggle_active_review_pane,
                "active": update_active_grade,
                "all": update_all_grades,
            }
        )


@contextmanager
def word():
    try:
        yield
    finally:
        docxrev.quit_word_safely()  # If used as a CLI, quit Word if nothing was open



"""Open all documents in preparation for grading."""

from __future__ import annotations

import docxrev
from win32com.client import constants

from gradedoc import shared


def open_all():
    """Open all documents in preparation for grading."""

    (paths, _) = shared.get_paths()
    for path in paths:
        document = docxrev.Document(path, save_on_exit=False, close_on_exit=False)
        with document:
            # Open the revisions pane.
            document.com.ActiveWindow.View.SplitSpecial = constants.wdPaneRevisions



"""Save and close all documents."""

from __future__ import annotations

import docxrev

from gradedoc import shared


def close_all():
    """Save and close all documents."""

    (paths, _) = shared.get_paths()
    for path in paths:
        document = docxrev.Document(path, save_on_exit=True, close_on_exit=True)
        with document:
            pass
